import { calculateDuration } from '../training/utils'
import { IExerciseItem } from './types'
import { TBlockInstance } from '../training/types'

/**
 * Find the right exercise given a certain time.
 * @param exerciseList
 * @param time
 */
export function getPosition(exerciseList: IExerciseItem[], time: number) {
  const index = exerciseList.findIndex((exercise) => time < exercise.offset + exercise.duration)

  const previousExercise = index >= 1 ? exerciseList[index - 1] : null
  const currentExercise = index >= 0 ? exerciseList[index] : null
  const nextExercise =
    index >= 0 && index < exerciseList.length - 1 ? exerciseList[index + 1] : null
  const values = {
    currentExercise,
    nextExercise,
    previousExercise,
    exerciseTime: 0,
  }
  if (currentExercise !== null) {
    values.exerciseTime = Math.max(time - currentExercise.offset, 0)
  }
  return values
}

/**
 * Takes a (nested) block of exercises and returns a flat array of exercises.
 * @param blocks
 * @param initialOffset - used for recursive application
 */
export function getExerciseList(
  blockInstances?: TBlockInstance[],
  initialOffset = 0,
  toplevelBlock: undefined | string = undefined
): IExerciseItem[] {
  if (!blockInstances?.length) return []
  let offset = initialOffset
  return blockInstances
    .map((blockInstance) => {
      const { block, rounds = 1 } = blockInstance
      if (block.blocks?.length) {
        const blockArray = []
        for (let i = 0; i < (rounds ?? 1); i++) {
          const subBlocks = getExerciseList(block.blocks, offset, toplevelBlock || block.title)
          const lastItem = subBlocks[subBlocks.length - 1]
          if (lastItem) {
            offset = lastItem.offset + lastItem.duration
          }
          if (block.rest) {
            if (lastItem && lastItem.exercise === 'Rest') {
              lastItem.duration += block.rest
              offset += block.rest
            } else {
              subBlocks.push({
                exercise: 'Rest',
                toplevelBlock: toplevelBlock || block.title,
                duration: block.rest,
                offset,
              })
              offset += block.rest
            }
          }
          blockArray.push(subBlocks)
        }
        return blockArray.flat()
      } else if (block.exercises) {
        console.log('found exercises', block.exercises)
        const blockArray: IExerciseItem[] = []
        const newItem = {
          exercise: block.exercises
            .map((exerciseInstance) => {
              const {
                exercise: { name },
                repetitions = 1,
              } = exerciseInstance
              return repetitions > 1 ? `${repetitions}x ${name}` : name
            })
            .join(' - '),
          duration: calculateDuration(blockInstance),
          videos: block.videos,
          description: block.description,
          toplevelBlock: toplevelBlock || block.title,
          offset,
        }
        for (let i = 0; i < (rounds ?? 1); i++) {
          blockArray.push({ ...newItem, offset })
          offset += newItem.duration
        }
        console.log(newItem, blockArray)
        return blockArray.flat()
      }
    })
    .flat()
}

export function getTrainingTime(exercises: IExerciseItem[]) {
  return exercises.reduce((accumulator, exercise) => accumulator + exercise.duration, 0)
}

export function polarToCartesian(
  centerX: number,
  centerY: number,
  radius: number,
  angleInDegrees: number
) {
  const angleInRadians = ((angleInDegrees - 90) * Math.PI) / 180.0

  return {
    x: centerX + radius * Math.cos(angleInRadians),
    y: centerY + radius * Math.sin(angleInRadians),
  }
}

export function describeArc(
  x: number,
  y: number,
  radius: number,
  startAngle: number,
  endAngle: number
) {
  const start = polarToCartesian(x, y, radius, endAngle)
  const end = polarToCartesian(x, y, radius, startAngle)

  const largeArcFlag = endAngle - startAngle <= 180 ? '0' : '1'

  const arcString = `M ${start.x} ${start.y} A ${radius} ${radius} 0 ${largeArcFlag} 0 ${end.x} ${end.y}`

  return arcString
}

export function limit(value: number, min: number, max: number) {
  return Math.min(Math.max(min, value), max)
}
